package BS;
use utf8;
# Bootstrap.pm
# © 2007-2017 Matías Software Group
# © 2007-2024 Salvador Ortiz García
#
# Bootstrap GranJefe.

use IO::File ();
use APR::Base64 ();
use YAML::Tiny 1.51 ();
use Tie::IxHash ();
use Encode ();
require bytes; # Needs to be outside VFS!? CHECK
require 'bytes_heavy.pl' if $bytes::VERSION < '1.09';

use version; our $VERSION = version->declare('v2.0.1');
$Bootstrap::VERSION = $VERSION;
our $lax;

BEGIN {
    if($GJ::InsideApache) {
        $GJ::Starting = 1;
        $GJ::MainServer->add_config(["PerlPassEnv $_"])
            for qw(GJ_SHOW_FAIL GJ_LAX_VERIFY MOD_PERL_GJDPATH MOD_PERL_GJUSER INVOCATION_ID);
        require Symbol;
        *GJ::gensym = \&Symbol::gensym;
    } else {
        # TODO: Any needed?
    }
    $lax = $ENV{GJ_LAX_VERIFY} ? '-noverify' : '';
    *B = \&APR::Base64::decode;
    eval qq|*{"\\x{@{[sprintf("%04x",$_+0x406)]}}"} =
       sub { pack '@{[chr($_+67)]}*', \@_ }| for(0,5,19);
    *ЙK = sub { [ Й(0x04034b50), Й(0x06054b50) ] };
    $GJ::VFS::Vfs = tie %GJ::VFS::VFS, 'Tie::IxHash';
    binmode(STDERR, ':utf8'); # So warn (and prejump Log) accept wide char
}

use strict;
sub ψ () { Ћ q<6f70656e73736c20736d696d65202d766572696679202573202d696e666f726d20444552202d434166696c65202f6574632f4772616e4a6566652f63612d62756e646c652e637274202d696e20257320323e2f6465762f6e756c6c> }
sub ѳ () { 'Archive::Zip' }

use overload
    '""' => sub { $_[0]->{path} },
    'cmp' => sub { "$_[0]" cmp $_[1] }
;

sub New { eval qq<require $_[0]; &$_[0]::new> };

sub Log {
    unshift @_, 'GJ' and goto &GJ::Log if defined &GJ::Log;
    shift if $_[0] eq 'notice';
    warn "@_\n";
}

sub Verify {
    my ($module, $handler, $pack) = @_;
    if ( -r $module and -s _ and -x _
        and $pack = sprintf(ψ, $lax, $module)
        and $module = qx($pack)
        and $pack = ЙK->[$?]
        and $handler = unpack('V', substr($module, 0, 4, $pack))
        and substr($module, $handler, 4, ЙK->[$?+1]) =~ /^GJ/
        and $pack = "$_[0]"
        and $_[0] = IO::GJ->new(\$module)
    ) {
        ${*{$_[0]}} = $pack;
        return 1; # Ok
    }
    return;
}

sub Register {
    my $fh = shift;
    my $storage = New ѳ;
    $storage->readFromFileHandle($fh);
    if(my $Manifest = $storage->contents('MANIFEST')) {
        $Manifest = YAML::Tiny->read_string($Manifest)->[0];
        my $fname = $fh->filename;
        my $self = bless {
            storage => $storage,
            name => (split('/', $fname))[-1],
            manifest => $Manifest,
            origin => $fname,
            lmt => (stat($fname))[9],
        }, 'GJ::VFS';
        $self->{mems} = {
            map {
                $_->{v} = $self;
                ($_->fileName => bless $_, 'VFS::ParMember')
            } $storage->members
        };
        Log('notice', "Binary $self->{name}:", Encode::decode_utf8($storage->zipfileComment));
        $GJ::VFS::Vfs->Push($self->{name} => $self);
        return $self;
    }
    return;
}

our %Preproc;
sub BS::INC {
    my $self = shift;
    my $file = shift;
    my ($fh, $fn);
    for(@{$self->{vfs}}) {
        my $m = $_->{manifest}{Modules} || $_->{manifest}{modules};
        last if $m->{$file} and $fn = $_->{mems}{"plib/$file"} and $fh = $fn->open();
    }
    unless($fh) { # Try real file
        $fn = "$self->{path}/$file" and -f $fn and $fh = IO::File->new($fn, '<');
    };
    if($fh) {
        my $prep = delete $Preproc{$file};
        $INC{$file} = $fn;
        return $prep ? ($fh, $prep->{filt}, $prep->{ctx}) : $fh;
    }
    return;
}

package IO::String;
package IO::GJ;
use parent -norequire => qw(IO::File IO::String);
sub new {
    my ($class, $strref) = @_;
    my $fh = $class->SUPER::new($strref, '+<:scalar');
    *{$fh}->{strref} = $strref;
    return $fh
}
sub filename { ${*{$_[0]}} };
sub string_ref { *{$_[0]}->{strref}; }
sub pos {
    my $self = shift;
    my $pos = $self->tell;
    $self->seek(shift, 0) if @_;
    return $pos;
}
package GJ::VFS;
use overload '""' => sub { "gjvfs:///$_[0]->{name}/" };
package VFS::Member;
use overload
    'bool' => sub { 1 },
    '""'   => sub { $_[0]->toString };
sub Container { $GJ::VFS::VFS{$_[0]->container}; }
sub toString { $_[0]->Container . $_[0]->fileName; }

package VFS::ParMember;
use parent -norequire => qw(Archive::Zip::ZipFileMember VFS::Member);
sub open {
    my $self = shift;
    my $fh;
    $fh = IO::File->new_tmpfile and $self->extractToFileHandle($fh) == 0
        or die "vfserr: $!\n";
    $fh->seek(0, 0);
    $fh->binmode(@_) if @_;
    return $fh;
}
sub container { $_[0]->{v}{name}; }
# All ready to install our INC hook
package BS;
{
    $GJ::_arch = `uname -m` and warn chop $GJ::_arch unless $GJ::_arch;
    Log('notice', "GranJefe Bootstrap $GJ::_arch $VERSION");
    for(@INC) {
        if(m{^(/usr/lib(?:\d+)?/GranJefe)/plib$}) {
            $GJ::_libdir = $1;
            unless(ref $_) {
                my @par;
                while(my $module = glob "$GJ::_libdir/plib/*.gjs") {
                    push @par, Register($module) if Verify($module);
                }
                $_ = bless { path => $_, sec => 1, vfs => \@par } if @par;
            }
            last;
        }
    }
    unless(eval { require GJ::Boot }) {
        warn "Fatal: Can't bootstrap!" . ($ENV{GJ_SHOW_FAIL} ? ": $@\n" : "\n");
        CORE::exit(1);
    }
}
