package GJ::Vite;
use utf8;
use warnings;
use strict;
use URL::Encode::XS();

use GJ::Boot 'VITE';
our $VERSION = '0.500';

sub Import {
    my $userclass = shift;
}

our %AgentGenCfg = (
    agent => 'GJAgent/1.0',
    timeout => 30,
    ssl_opts => {
        verify_hostname => 1,
        # For debug certificate validity errors
        #SSL_verify_callback => sub { warn "CB: @_[0, 3]\n"; return $_[0]; }
    }
);

our %REPO_METAS = (
    generic => {
        protocol => 'https://',
        repo => 'packages/generic/',
        extra => \'${_a[0]}/${_a[1]}/${_a[0]}.zip',
    },
    npm => => {
        protocol => 'npm://',
        repo => 'packages/npm/',
        extra => '',
    },
    artifacts => {
        protocol => 'https://',
        repo => 'jobs/artifacts/',
        extra => \'${UE(_a[0])}/download?job=${_a[1]}',
    }
);

sub _repotipe {
    my ($tpl, $rct, $var, $pid, @args) = &GJ::Template::varargs;
    my $meta = $REPO_METAS{$var};
    $tpl->assign($meta);
    $tpl->assign(proj => $pid);
    $tpl->assign(_a => \@args);
    return \q|${include 'BaseRepo'}${include extra}|;
}

sub _upgrade {
    my $self = shift;
    my $mod = shift;
    my $tpl = GJ::Template::Engine->new(
        module => $mod,
        valuer => sub {
            my($t, $var) = @_;
            local $t->{strict} = 0;
            my $val = $ENV{$var};
            $t->default($var => $val);
            return $val;
        },
        default => {
            UE => sub { URL::Encode::XS::url_encode($_[3]) },
            map { $_ => \&_repotipe } keys %REPO_METAS,
        },
        templates => { }
    );
    while(my ($ky,$v) = each %$self) {
        $tpl->addFromData($ky, $v);
    }
    $self->{_t} = $tpl;
    $self->{mb} = $mod->MagicBase;
    bless $self;
}

sub Init {
    my $Module = shift;
    my $self = shift || {};
    my %activate;
    if('HASH' eq ref $self) {
        my $exts = delete $self->{Externals};
        _upgrade($self, $Module);
        while (my ($key, $conf) = each %{$exts})  {
            if('HASH' eq ref $conf) {
                if(my $tok = GJ::Vite::Ext->new($self, $key, $conf)) {
                    $Module->Log(info => "GJVite: Will handle '$key' with '$tok'");
                    $activate{$key} = "$tok/"; # Avoid one redirection round trip
                }
            } else {
                $Module->Log(warn => "GJVite: ignored '$key' malformed  '$conf'!")
            }
        }
    } else {
        $Module->Log(warn => "GJVite: malformed  $self!")
    }
    GJ::FWServices::PostDispatchReg($Module, sub {
        # Tobe executed once per request after Module's Work
        my($Module, $r, $action) = @_;
        my $res = $action;
        unless($Module->Attended || ref($action) ||
               Scalar::Util::looks_like_number($res)
        ) {
            my $tok = $activate{$action};
            if($tok) {
                warn "V: $action $tok\n";
                $Module->SendRedirect($activate{$action})
            }
        }
        return undef;
    });
    1;
}

package GJ::Vite::Ext;

sub new {
    my $proto = shift;
    my $top = shift;
    my $key = shift;
    my $ext = shift;
    my $Module = $top->{_t}->Module;
    unless($ext->{Path}) {
        $Module->Log(warn => "Vite: No path in $key conf");
        return
    }
    $ext->{Mode} ||= 'auto';
    $ext->{key} = $key;
    bless $ext, $proto;
    $ext->analize($top, $Module);

    my $path = $ext->{Path};
    my $mode = $ext->{Mode};
    my $mbase = $top->{mb};
    my $base = $ext->{Base} || $Module->Param("Vite_${key}_Base") || $key;
    my $intro = $ext->{Intro} || '';
    $base = "/$base" unless $base =~ /^\//;
    $Module->Log(info => "GJVite: Connect '$key' $base $path");
    if($mode eq 'static') {
        # Prepare static content in $base
        $Module->Log(info => "GJVite: Configuring static $key...");
        if(!-f "$path/index.html") {
            $Module->Log(warn => "No index.html in '$key'");
        }
        $Module->AddApacheConf( qq|Alias $base $path| );
        return $base . $intro;
    } elsif($mode eq 'proxy') {
        $Module->Log(info => "GJVite: Configuring proxy $key");
        my $port = $ext->{Port} || $Module->Param("Vite_${key}_Port") || 4173;
        my $selfsvr = "localhost:80";
        $selfsvr =~ s/:\d+/':' . $port/e;
        $Module->Log(info => "GJVite PB: $selfsvr/$base");
        $Module->AddApacheConf(
            qq|ProxyPass "/$base" "http://$selfsvr$base" upgrade=websocket|
        );
        return $base . $intro; # No ready yet
    } else {
        $Module->Log(warn => "Vite: unknow mode $mode, ignored $base")
    }
    return undef;
}


sub analize {
    my ($self, $top, $mod) = @_;
    $_ = $self->{Path};
    unless(/^\//) {
        # Make absolute...
        my $p;
        if(!tr|/|/| && (my $local = $mod->OnLibDir($_))) {
            $_ = "$local"; # realfile
        } elsif($mod->{_datadir} && ($p = "$mod->{_datadir}/$_") && -d $p) {
            $_ = $p;
        } elsif($p = "$GJ::ServerRoot/lib/externals/$_") {
            if(-d $p) { $_ = $p }
            elsif($self->{Create}) {
                $mod->Log('notice', "Creating $p");
                if(system("mkdir -p -m 0750 $p") || system("chown $GJ::Privs $p")) {
                    $mod->Log('error', "Can't create $p: $?");
                    return;
                }
                IO::File->new("$p/.populate_me", '>') or return;
                $_ = $p;
            }
        } else {
            die "Vite: can't locate extenal $_";
        }
    }
    if(-f "$_/.populate_me") {
        my $tpl = $top->{_t};
        my $url = join('', $tpl->subparse(shift @{$self->{Create}}));
        $tpl->addFromData(Create => $self->{Create});
        $mod->Log(info => "Downloading: $url");
        if($url =~ /^http/) {
            require LWP::UserAgent;
            my $exdir = $_;
            my $zip = '/tmp/result.zip';
            my @args = (':content_file' => $zip, $tpl->expand_dr('Create'));
            my $ua = $top->{ua} ||= LWP::UserAgent->new(%AgentGenCfg);
            my $response = $ua->get($url, @args);
            if($response->is_success) {
                system('unzip', $zip, '-d', $exdir);
            } else {
                die "Download error:", $response->status_line
            }
        }
        unlink "$_/.populate_me"
    }
    if(-d "$_/dist") {
        $self->{Type} = 'svelte';
        $self->{Mode} = 'static';
        $self->{Path} = "$_/dist"; 
    }
    elsif(-d "$_/build") {
        $self->{Type} = 'kit';
        $self->{Mode} = 'static';
        $self->{Path} = "$_/build"; 
    }
    if($self->{Type}) {
        $mod->Log(info => "Type of $self->{key}: $self->{Type}");
    }
}

